'''
XbrlPublicPostgresDB.py implements a relational database interface for Arelle, based
on the XBRL US Public Database.  The database schema is described by "XBRL US Public Database",
published by XBRL US, 2011.  This is a syntactic representation of XBRL information.

This module provides the execution context for saving a dts and assession in
XBRL Public Database Tables.  It may be loaded by Arelle'sRSS feed, or by individual
DTS and instances opened by interactive or command line/web service mode.

See COPYRIGHT.md for copyright information.
(Does not apply to the XBRL US Database schema and description.)

The XBRL US Database schema and description is (c) Copyright XBRL US 2011, The
resulting database may contain data from SEC interactive data filings (or any other XBRL
instance documents and DTS) in a relational model. Workiva, Inc. conveys neither
rights nor license for the database schema.

The XBRL US Database and this code is intended for Postgres.  XBRL-US uses Postgres 8.4,
Arelle uses 9.1, via Python DB API 2.0 interface, using the pg8000 library.

Information for the 'official' XBRL US-maintained database (this schema, containing SEC filings):
    Database Name: edgar_db
    Database engine: Postgres version 8.4
    Host: public.xbrl.us
    Port: 5432

to use from command line:

linux
   # be sure plugin is installed
   arelleCmdLine --plugin '+xbrlDB|show'
   arelleCmdLine -f http://sec.org/somewhere/some.rss -v --store-to-XBRL-DB 'myserver.com,portnumber,pguser,pgpasswd,database,timeoutseconds'

windows
   rem be sure plugin is installed
   arelleCmdLine --plugin "+xbrlDB|show"
   arelleCmdLine -f http://sec.org/somewhere/some.rss -v --store-to-XBRL-DB "myserver.com,portnumber,pguser,pgpasswd,database,timeoutseconds"

'''

import os, time, datetime
from arelle.ModelDocument import Type
from arelle.ModelDtsObject import ModelConcept, ModelResource
from arelle.ModelObject import ModelObject
from arelle.ModelValue import qname
from arelle.ValidateXbrlCalcs import roundValue
from arelle.XmlUtil import elementFragmentIdentifier
from arelle import XbrlConst
from .SqlDb import XPDBException, isSqlConnection, SqlDbConnection
from arelle.PythonUtil import STR_NUM_TYPES


def insertIntoDB(modelXbrl,
                 user=None, password=None, host=None, port=None, database=None, timeout=None,
                 product="postgres", rssItem=None, **kwargs):
    xpgdb = None
    try:
        xpgdb = XbrlPostgresDatabaseConnection(modelXbrl, user, password, host, port, database, timeout, product)
        xpgdb.verifyTables()
        xpgdb.insertXbrl(rssItem=rssItem)
        xpgdb.close()
    except Exception as ex:
        if xpgdb is not None:
            try:
                xpgdb.close(rollback=True)
            except Exception as ex2:
                pass
        raise # reraise original exception with original traceback

def isDBPort(host, port, timeout=10):
    return isSqlConnection(host, port, timeout, product="postgres")

XBRLDBTABLES = {
                "fact", "fact_aug",
                "entity",
                "entity_name_history",
                "unit", "unit_measure",
                "context", "context_aug", "context_dimension",
                "accession", "accession_document_association", "accession_element", "accession_timestamp",
                "attribute_value",
                "custom_role_type",
                "uri",
                "document",
                "qname",
                "taxonomy", "taxonomy_version", "namespace",
                "element", "element_attribute", "element_attribute_value_association",
                "network", "relationship",
                "custom_arcrole_type", "custom_arcrole_used_on", "custom_role_used_on",
                "label_resource",
                "reference_part", "reference_part_type", "reference_resource",
                "resource",
                "enumeration_arcrole_cycles_allowed",
                "enumeration_element_balance",
                "enumeration_element_period_type",
                "enumeration_unit_measure_location",
                "industry", "industry_level",
                "industry_structure",
                "query_log",
                "sic_code",
                }



class XbrlPostgresDatabaseConnection(SqlDbConnection):
    def verifyTables(self):
        missingTables = XBRLDBTABLES - self.tablesInDB()
        # if no tables, initialize database
        if missingTables == XBRLDBTABLES:
            self.create(os.path.join("sql", "public", "xbrlPublicPostgresDB.ddl"))

            # load fixed tables
            self.getTable('enumeration_arcrole_cycles_allowed', 'enumeration_arcrole_cycles_allowed_id',
                          ('description',), ('description',),
                          (('any',), ('undirected',), ('none',)))
            self.getTable('enumeration_element_balance', 'enumeration_element_balance_id',
                          ('description',), ('description',),
                          (('credit',), ('debit',)))
            self.getTable('enumeration_element_period_type', 'enumeration_element_period_type_id',
                          ('description',), ('description',),
                          (('instant',), ('duration',), ('forever',)))
            missingTables = XBRLDBTABLES - self.tablesInDB()
        if missingTables:
            raise XPDBException("xpgDB:MissingTables",
                                _("The following tables are missing, suggest reinitializing database schema: %(missingTableNames)s"),
                                missingTableNames=', '.join(t for t in sorted(missingTables)))

    def insertXbrl(self, rssItem):
        try:
            # must also have default dimensions loaded
            from arelle import ValidateXbrlDimensions
            ValidateXbrlDimensions.loadDimensionDefaults(self.modelXbrl)

            # find pre-existing documents in server database
            self.identifyPreexistingDocuments()
            self.identifyConceptsUsed()

            startedAt = time.time()
            self.insertAccession(rssItem)
            self.insertUris()
            self.insertQnames()
            self.insertNamespaces()
            self.insertDocuments()
            self.insertCustomArcroles()
            self.insertCustomRoles()
            self.insertElements()
            self.insertResources()
            self.insertNetworks()
            self.modelXbrl.profileStat(_("XbrlPublicDB: DTS insertion"), time.time() - startedAt)
            startedAt = time.time()
            self.insertFacts()
            self.modelXbrl.profileStat(_("XbrlPublicDB: instance insertion"), time.time() - startedAt)
            startedAt = time.time()
            self.showStatus("Committing entries")
            self.commit()
            self.modelXbrl.profileStat(_("XbrlPublicDB: insertion committed"), time.time() - startedAt)
            self.showStatus("DB insertion completed", clearAfter=5000)
        except Exception as ex:
            self.showStatus("DB insertion failed due to exception", clearAfter=5000)
            raise

    def insertAccession(self, rssItem):
        self.accessionId = "(TBD)"
        self.showStatus("insert accession")
        if rssItem is None:
            _time = time.time()
            now = datetime.datetime.fromtimestamp(_time)
            today = datetime.date(now.year, now.month, now.day)
            table = self.getTable('accession', 'accession_id',
                                  ('filing_date','entity_id','creation_software',
                                   'entry_url', 'filing_accession_number'),
                                  ('filing_accession_number',),
                                  ((today,  # NOT NULL
                                    0,  # NOT NULL
                                    self.modelXbrl.modelDocument.creationSoftwareComment,
                                    self.modelXbrl.uri,
                                    str(int(_time))  # NOT NULL
                                    ),),
                                  checkIfExisting=True,
                                  returnExistenceStatus=True)
        else:
            table = self.getTable('accession', 'accession_id',
                                  ('accepted_timestamp', 'is_most_current', 'filing_date','entity_id',
                                   'entity_name', 'creation_software', 'standard_industrial_classification',
                                   'sec_html_url', 'entry_url', 'filing_accession_number'),
                                  ('filing_accession_number',),
                                  ((rssItem.acceptanceDatetime,
                                    True,
                                    rssItem.filingDate or datetime.datetime.min,  # NOT NULL
                                    rssItem.cikNumber or 0,  # NOT NULL
                                    rssItem.companyName,
                                    self.modelXbrl.modelDocument.creationSoftwareComment,
                                    rssItem.assignedSic or -1,  # NOT NULL
                                    rssItem.htmlUrl,
                                    rssItem.url,
                                    rssItem.accessionNumber or 'UNKNOWN'  # NOT NULL
                                    ),),
                                  checkIfExisting=True,
                                  returnExistenceStatus=True)
        for id, filing_accession_number, existenceStatus in table:
            self.accessionId = id
            self.accessionPreviouslyInDB = existenceStatus
            break

    def insertUris(self):
        uris = (self.modelXbrl.namespaceDocs.keys() |
                self.modelXbrl.arcroleTypes.keys() |
                XbrlConst.standardArcroleCyclesAllowed.keys() |
                self.modelXbrl.roleTypes.keys() |
                XbrlConst.standardRoles)
        self.showStatus("insert uris")
        table = self.getTable('uri', 'uri_id',
                              ('uri',),
                              ('uri',), # indexed match cols
                              tuple((uri,)
                                    for uri in uris),
                              checkIfExisting=True)
        self.uriId = dict((uri, id)
                          for id, uri in table)

    def insertQnames(self):
        qnames = (self.modelXbrl.qnameConcepts.keys() |
                  self.modelXbrl.qnameAttributes.keys() |
                  self.modelXbrl.qnameTypes.keys() |
                  set(measure
                      for unit in self.modelXbrl.units.values()
                      for measures in unit.measures
                      for measure in measures))
        self.showStatus("insert qnames")
        table = self.getTable('qname', 'qname_id',
                              ('namespace', 'local_name'),
                              ('namespace', 'local_name'), # indexed match cols
                              tuple((qn.namespaceURI, qn.localName)
                                    for qn in qnames),
                              checkIfExisting=True)
        self.qnameId = dict((qname(ns, ln), id)
                            for id, ns, ln in table)

    def insertNamespaces(self):
        self.showStatus("insert namespaces")
        if self.disclosureSystem.baseTaxonomyNamespaces:
            # use only base taxonomy namespaces in disclosure system
            namespaceUris = self.disclosureSystem.baseTaxonomyNamespaces
        else:
            # use all namespace URIs
            namespaceUris = self.modelXbrl.namespaceDocs.keys()
        table = self.getTable('namespace', 'namespace_id',
                              ('uri', 'is_base', 'taxonomy_version_id', 'prefix'),
                              ('uri',), # indexed matchcol
                              tuple((uri, True, 0, self.disclosureSystem.standardPrefixes.get(uri,None))
                                    for uri in namespaceUris),
                              checkIfExisting=True)
        self.namespaceId = dict((uri, id)
                                for id, uri in table)

    def identifyPreexistingDocuments(self):
        self.existingDocumentIds = {}
        docUris = set()
        for modelDocument in self.modelXbrl.urlDocs.values():
            if modelDocument.type == Type.SCHEMA:
                docUris.add(self.dbStr(modelDocument.uri))
        if docUris:
            results = self.execute("SELECT document_id, document_uri FROM document WHERE document_uri IN (" +
                                   ', '.join(docUris) + ");")
            self.existingDocumentIds = dict((docUri,docId) for docId, docUri in results)

    def identifyConceptsUsed(self):
        # relationshipSets are a dts property
        self.relationshipSets = [(arcrole, ELR, linkqname, arcqname)
                                 for arcrole, ELR, linkqname, arcqname in self.modelXbrl.baseSets.keys()
                                 if ELR and (arcrole.startswith("XBRL-") or (linkqname and arcqname))]


        conceptsUsed = set(f.qname for f in self.modelXbrl.factsInInstance)

        for cntx in self.modelXbrl.contexts.values():
            for dim in cntx.qnameDims.values():
                conceptsUsed.add(dim.dimensionQname)
                if dim.isExplicit:
                    conceptsUsed.add(dim.memberQname)
                else:
                    conceptsUsed.add(dim.typedMember.qname)
        for defaultDim, defaultDimMember in self.modelXbrl.qnameDimensionDefaults.items():
            conceptsUsed.add(defaultDim)
            conceptsUsed.add(defaultDimMember)
        for relationshipSetKey in self.relationshipSets:
            relationshipSet = self.modelXbrl.relationshipSet(*relationshipSetKey)
            for rel in relationshipSet.modelRelationships:
                if isinstance(rel.fromModelObject, ModelConcept):
                    conceptsUsed.add(rel.fromModelObject)
                if isinstance(rel.toModelObject, ModelConcept):
                    conceptsUsed.add(rel.toModelObject)
        for qn in (XbrlConst.qnXbrliIdentifier, XbrlConst.qnXbrliPeriod, XbrlConst.qnXbrliUnit):
            conceptsUsed.add(self.modelXbrl.qnameConcepts[qn])

        conceptsUsed -= {None}  # remove None if in conceptsUsed
        self.conceptsUsed = conceptsUsed

    def insertDocuments(self):
        self.showStatus("insert documents")
        table = self.getTable('document', 'document_id',
                              ('document_uri',),
                              ('document_uri',),
                              set((docUri,)
                                  for docUri in self.modelXbrl.urlDocs.keys()
                                  if docUri not in self.existingDocumentIds),
                              checkIfExisting=True)
        self.documentIds = dict((uri, id)
                                for id, uri in table)
        self.documentIds.update(self.existingDocumentIds)
        table = self.getTable('accession_document_association', 'accession_document_association_id',
                              ('accession_id','document_id'),
                              ('document_id',),
                              tuple((self.accessionId, docId)
                                    for docId in self.documentIds.values()),
                              checkIfExisting=True)

    def insertCustomArcroles(self):
        self.showStatus("insert arcrole types")
        arcroleTypesByIds = dict(((self.documentIds[arcroleType.modelDocument.uri],
                                   self.uriId[arcroleType.arcroleURI]), # key on docId, uriId
                                  arcroleType) # value is roleType object
                                 for arcroleTypes in self.modelXbrl.arcroleTypes.values()
                                 for arcroleType in arcroleTypes
                                 if arcroleType.modelDocument.uri not in self.existingDocumentIds)
        table = self.getTable('custom_arcrole_type', 'custom_arcrole_type_id',
                              ('document_id', 'uri_id', 'definition', 'cycles_allowed'),
                              ('document_id', 'uri_id'),
                              tuple((arcroleTypeIDs[0], # doc Id
                                     arcroleTypeIDs[1], # uri Id
                                     arcroleType.definition,
                                     {'any':1, 'undirected':2, 'none':3}[arcroleType.cyclesAllowed])
                                    for arcroleTypeIDs, arcroleType in arcroleTypesByIds.items()))
        table = self.getTable('custom_arcrole_used_on', 'custom_arcrole_used_on_id',
                              ('custom_arcrole_type_id', 'qname_id'),
                              ('custom_arcrole_type_id', 'qname_id'),
                              tuple((id, self.qnameId[usedOn])
                                    for id, docid, uriid in table
                                    for usedOn in arcroleTypesByIds[(docid,uriid)].usedOns))

    def insertCustomRoles(self):
        self.showStatus("insert role types")
        roleTypesByIds = dict(((self.documentIds[roleType.modelDocument.uri],
                                self.uriId[roleType.roleURI]), # key on docId, uriId
                               roleType) # value is roleType object
                              for roleTypes in self.modelXbrl.roleTypes.values()
                              for roleType in roleTypes
                              if roleType.modelDocument.uri not in self.existingDocumentIds)
        table = self.getTable('custom_role_type', 'custom_role_type_id',
                              ('document_id', 'uri_id', 'definition'),
                              ('document_id', 'uri_id'),
                              tuple((roleTypeIDs[0], # doc Id
                                     roleTypeIDs[1], # uri Id
                                     roleType.definition)
                                    for roleTypeIDs, roleType in roleTypesByIds.items()))
        table = self.getTable('custom_role_used_on', 'custom_role_used_on_id',
                              ('custom_role_type_id', 'qname_id'),
                              ('custom_role_type_id', 'qname_id'),
                              tuple((id, self.qnameId[usedOn])
                                    for id, docid, uriid in table
                                    for usedOn in roleTypesByIds[(docid,uriid)].usedOns))

    def insertElements(self):
        self.showStatus("insert elements")

        filingDocumentConcepts = set()
        existingDocumentUsedConcepts = set()
        for concept in self.modelXbrl.qnameConcepts.values():
            if concept.modelDocument.uri not in self.existingDocumentIds:
                filingDocumentConcepts.add(concept)
            elif concept in self.conceptsUsed:
                existingDocumentUsedConcepts.add(concept)

        table = self.getTable('element', 'element_id',
                              ('qname_id', 'datatype_qname_id', 'xbrl_base_datatype_qname_id', 'balance_id',
                               'period_type_id', 'substitution_group_qname_id', 'abstract', 'nillable',
                               'document_id', 'is_numeric', 'is_monetary'),
                              ('qname_id',),
                              tuple((self.qnameId[concept.qname],
                                     self.qnameId.get(concept.typeQname), # may be None
                                     self.qnameId.get(concept.baseXbrliTypeQname
                                                      if not isinstance(concept.baseXbrliTypeQname, list)
                                                      else concept.baseXbrliTypeQname[0]
                                                      ), # may be None or may be a list for a union
                                     {'debit':1, 'credit':2, None:None}[concept.balance],
                                     {'instant':1, 'duration':2, 'forever':3, None:0}[concept.periodType],
                                     self.qnameId.get(concept.substitutionGroupQname), # may be None
                                     concept.isAbstract,
                                     concept.isNillable,
                                     self.documentIds[concept.modelDocument.uri],
                                     concept.isNumeric,
                                     concept.isMonetary)
                                    for concept in filingDocumentConcepts)
                             )
        self.elementId = dict((qnameId, elementId)  # indexed by qnameId, not by qname value
                              for elementId, qnameId in table)
        filingDocumentConcepts.clear() # dereference

        # get existing element IDs
        if existingDocumentUsedConcepts:
            conceptQnameIds = []
            for concept in existingDocumentUsedConcepts:
                conceptQnameIds.append(str(self.qnameId[concept.qname]))
            results = self.execute("SELECT element_id, qname_id FROM element WHERE qname_id IN (" +
                                   ', '.join(conceptQnameIds) + ");")
            for elementId, qnameId in results:
                self.elementId[qnameId] = elementId
        existingDocumentUsedConcepts.clear() # dereference

    def conceptElementId(self, concept):
        if isinstance(concept, ModelConcept):
            return self.elementId.get(self.qnameId.get(concept.qname))
        else:
            return None

    def insertResources(self):
        self.showStatus("insert resources")
        # deduplicate resources (may be on multiple arcs)
        # note that lxml has no column numbers, use objectIndex as pseudo-column number
        uniqueResources = dict(((self.documentIds[resource.modelDocument.uri],
                                 resource.sourceline,
                                 resource.objectIndex), resource)
                               for arcrole in (XbrlConst.conceptLabel, XbrlConst.conceptReference,
                                               XbrlConst.elementLabel, XbrlConst.elementReference)
                               for rel in self.modelXbrl.relationshipSet(arcrole).modelRelationships
                               for resource in (rel.fromModelObject, rel.toModelObject)
                               if isinstance(resource, ModelResource))
        resourceData = tuple((self.uriId[resource.role],
                              self.qnameId[resource.qname],
                              self.documentIds[resource.modelDocument.uri],
                              resource.sourceline,
                              resource.objectIndex)
                             for resource in uniqueResources.values())
        uniqueResources.clear() # dereference before getTable
        table = self.getTable('resource', 'resource_id',
                              ('role_uri_id', 'qname_id', 'document_id', 'document_line_number', 'document_column_number'),
                              ('document_id', 'document_line_number', 'document_column_number'),
                              resourceData,
                              checkIfExisting=True)
        self.resourceId = dict(((docId, line, offset), id)
                               for id, docId, line, offset in table)

        self.showStatus("insert labels")
        uniqueResources = dict(((self.resourceId[self.documentIds[resource.modelDocument.uri],
                                                     resource.sourceline,
                                                     resource.objectIndex]), resource)
                               for arcrole in (XbrlConst.conceptLabel, XbrlConst.elementLabel)
                               for rel in self.modelXbrl.relationshipSet(arcrole).modelRelationships
                               for resource in (rel.fromModelObject, rel.toModelObject)
                               if isinstance(resource, ModelResource))
        table = self.getTable('label_resource', 'resource_id',
                              ('resource_id', 'label', 'xml_lang'),
                              ('resource_id',),
                              tuple((resourceId,
                                     resource.stringValue,
                                     resource.xmlLang)
                                    for resourceId, resource in uniqueResources.items()),
                              checkIfExisting=True)
        uniqueResources.clear()

    def insertNetworks(self):
        self.showStatus("insert networks")
        table = self.getTable('network', 'network_id',
                              ('accession_id', 'extended_link_qname_id', 'extended_link_role_uri_id',
                               'arc_qname_id', 'arcrole_uri_id', 'description'),
                              ('accession_id', 'extended_link_qname_id', 'extended_link_role_uri_id',
                               'arc_qname_id', 'arcrole_uri_id'),
                              tuple((self.accessionId,
                                     self.qnameId[linkqname],
                                     self.uriId[ELR],
                                     self.qnameId[arcqname],
                                     self.uriId[arcrole],
                                     None if ELR in XbrlConst.standardRoles else
                                     self.modelXbrl.roleTypes[ELR][0].definition)
                                    for arcrole, ELR, linkqname, arcqname in self.modelXbrl.baseSets.keys()
                                    if ELR and linkqname and arcqname and not arcrole.startswith("XBRL-")))
        self.networkId = dict(((accId, linkQnId, linkRoleId, arcQnId, arcRoleId), id)
                              for id, accId, linkQnId, linkRoleId, arcQnId, arcRoleId in table)
        # do tree walk to build relationships with depth annotated, no targetRole navigation
        dbRels = []

        def walkTree(rels, seq, depth, relationshipSet, visited, dbRels, networkId):
            for rel in rels:
                if rel not in visited and isinstance(rel.toModelObject, ModelObject):
                    visited.add(rel)
                    dbRels.append((rel, seq, depth, networkId))
                    seq += 1
                    seq = walkTree(relationshipSet.fromModelObject(rel.toModelObject), seq, depth+1, relationshipSet, visited, dbRels, networkId)
                    visited.remove(rel)
            return seq

        for arcrole, ELR, linkqname, arcqname in self.modelXbrl.baseSets.keys():
            if ELR and linkqname and arcqname and not arcrole.startswith("XBRL-"):
                networkId = self.networkId[(self.accessionId,
                                            self.qnameId[linkqname],
                                            self.uriId[ELR],
                                            self.qnameId[arcqname],
                                            self.uriId[arcrole])]
                relationshipSet = self.modelXbrl.relationshipSet(arcrole, ELR, linkqname, arcqname)
                seq = 1
                for rootConcept in relationshipSet.rootConcepts:
                    seq = walkTree(relationshipSet.fromModelObject(rootConcept), seq, 1, relationshipSet, set(), dbRels, networkId)

        def resourceResourceId(resource):
            if isinstance(resource, ModelResource):
                return self.resourceId.get((self.documentIds[resource.modelDocument.uri],
                                            resource.sourceline,
                                            resource.objectIndex))
            else:
                return None

        relsData = tuple((networkId,
                          self.conceptElementId(rel.fromModelObject), # may be None
                          self.conceptElementId(rel.toModelObject), # may be None
                          self.dbNum(rel.order),
                          resourceResourceId(rel.fromModelObject), # may be None
                          resourceResourceId(rel.toModelObject), # may be None
                          self.dbNum(rel.weight), # none if no weight
                          sequence,
                          depth,
                          self.qnameId.get(rel.preferredLabel) if rel.preferredLabel else None)
                         for rel, sequence, depth, networkId in dbRels
                         if isinstance(rel.fromModelObject, ModelConcept) and isinstance(rel.toModelObject, ModelConcept))
        del dbRels[:]   # dererefence
        table = self.getTable('relationship', 'relationship_id',
                              ('network_id', 'from_element_id', 'to_element_id', 'reln_order',
                               'from_resource_id', 'to_resource_id', 'calculation_weight',
                               'tree_sequence', 'tree_depth', 'preferred_label_role_uri_id'),
                              ('network_id', 'tree_sequence'),
                              relsData)

    def insertFacts(self):
        accsId = self.accessionId
        if self.accessionPreviouslyInDB:
            self.showStatus("deleting prior facts of this accession")
            # remove prior facts
            self.execute("DELETE FROM fact WHERE fact.accession_id = {};".format(accsId),
                         close=False, fetch=False)
            self.execute("DELETE from unit_measure "
                         "USING unit "
                         "WHERE unit.accession_id = {0} AND unit_measure.unit_id = unit.unit_id;".format(accsId),
                         close=False, fetch=False)
            self.execute("DELETE from unit WHERE unit.accession_id = {0};".format(accsId),
                         close=False, fetch=False)
            self.execute("DELETE from context WHERE context.accession_id = {0};".format(accsId),
                         close=False, fetch=False)
        self.showStatus("insert facts")
        # units
        table = self.getTable('unit', 'unit_id',
                              ('accession_id', 'unit_xml_id'),
                              ('accession_id', 'unit_xml_id'),
                              tuple((accsId,
                                     unitId)
                                    for unitId in self.modelXbrl.units.keys()))
        self.unitId = dict(((_accsId, xmlId), id)
                           for id, _accsId, xmlId in table)
        # measures
        table = self.getTable('unit_measure', 'unit_measure_id',
                              ('unit_id', 'qname_id', 'location_id'),
                              ('qname_id', 'location_id'),
                              tuple((self.unitId[(accsId,unit.id)],
                                     self.qnameId[measure],
                                     1 if (not unit.measures[1]) else (i + 1))
                                    for unit in self.modelXbrl.units.values()
                                    for i in range(2)
                                    for measure in unit.measures[i]))
        #table = self.getTable('enumeration_measure_location', 'enumeration_measure_location_id',
        #                      ('description',),
        #                      ('description',),
        #                      (('measure',), ('numerator',), ('denominator',)))
        # context
        table = self.getTable('context', 'context_id',
                              ('accession_id', 'period_start', 'period_end', 'period_instant', 'specifies_dimensions', 'context_xml_id', 'entity_scheme', 'entity_identifier'),
                              ('accession_id', 'context_xml_id'),
                              tuple((accsId,
                                     cntx.startDatetime if cntx.isStartEndPeriod else None,
                                     cntx.endDatetime if cntx.isStartEndPeriod else None,
                                     cntx.instantDatetime if cntx.isInstantPeriod else None,
                                     bool(cntx.qnameDims),
                                     cntx.id,
                                     cntx.entityIdentifier[0],
                                     cntx.entityIdentifier[1])
                                    for cntx in self.modelXbrl.contexts.values()))
        self.cntxId = dict(((_accsId, xmlId), id)
                           for id, _accsId, xmlId in table)
        # context_dimension
        values = []
        for cntx in self.modelXbrl.contexts.values():
            for dim in cntx.qnameDims.values():
                values.append((self.cntxId[(accsId,cntx.id)],
                               self.qnameId[dim.dimensionQname],
                               self.qnameId.get(dim.memberQname), # may be None
                               self.qnameId.get(dim.typedMember.qname) if dim.isTyped else None,
                               False, # not default
                               dim.contextElement == "segment",
                               dim.typedMember.stringValue if dim.isTyped else None))
            for dimQname, memQname in self.modelXbrl.qnameDimensionDefaults.items():
                if dimQname not in cntx.qnameDims:
                    values.append((self.cntxId[(accsId,cntx.id)],
                                   self.qnameId[dimQname],
                                   self.qnameId[memQname],
                                   None,
                                   True, # is default
                                   True, # ambiguous and irrelevant for the XDT model
                                   None))
        table = self.getTable('context_dimension', 'context_dimension_id',
                              ('context_id', 'dimension_qname_id', 'member_qname_id', 'typed_qname_id', 'is_default', 'is_segment', 'typed_text_content'),
                              ('context_id', 'dimension_qname_id', 'member_qname_id'), # shouldn't typed_qname_id be here?  not good idea because it's not indexed in XBRL-US DDL
                              values)
        # facts
        def insertFactSet(modelFacts, tupleFactId):
            table = self.getTable('fact', 'fact_id',
                                  ('accession_id', 'tuple_fact_id', 'context_id', 'unit_id', 'element_id', 'effective_value', 'fact_value',
                                   'xml_id', 'precision_value', 'decimals_value',
                                   'is_precision_infinity', 'is_decimals_infinity', ),
                                  ('accession_id', 'xml_id'),
                                  tuple((accsId,
                                         tupleFactId,
                                         self.cntxId.get((accsId,fact.contextID)),
                                         self.unitId.get((accsId,fact.unitID)),
                                         self.conceptElementId(fact.concept),
                                         roundValue(fact.value, fact.precision, fact.decimals) if fact.isNumeric and not fact.isNil else None,
                                         fact.value,
                                         elementFragmentIdentifier(fact),
                                         fact.xAttributes['precision'].xValue if ('precision' in fact.xAttributes and isinstance(fact.xAttributes['precision'].xValue,int)) else None,
                                         fact.xAttributes['decimals'].xValue if ('decimals' in fact.xAttributes and isinstance(fact.xAttributes['decimals'].xValue,int)) else None,
                                         'precision' in fact.xAttributes and fact.xAttributes['precision'].xValue == 'INF',
                                         'decimals' in fact.xAttributes and fact.xAttributes['decimals'].xValue == 'INF',
                                         )
                                        for fact in modelFacts))
            factId = dict((xmlId, id)
                          for id, _accsId, xmlId in table)
            for fact in modelFacts:
                if fact.isTuple:
                    insertFactSet(fact.modelTupleFacts,
                                  factId[elementFragmentIdentifier(fact)])
        insertFactSet(self.modelXbrl.facts, None)
        # hashes



'''
See COPYRIGHT.md for copyright information.
'''
from tkinter import Toplevel, N, S, E, W
try:
    from tkinter.ttk import Frame, Button
except ImportError:
    from ttk import Frame, Button
import regex as re
from arelle.UiUtil import (gridHdr, gridCell, gridCombobox, label, checkbox)

'''
caller checks accepted, if True, caller retrieves url
'''
def getParameters(mainWin):
    dialog = DialogFormulaParameters(mainWin, mainWin.modelManager.formulaOptions.__dict__.copy())
    if dialog.accepted:
        mainWin.modelManager.formulaOptions.__dict__.update(dialog.options)
        mainWin.config["formulaParameters"] = dialog.options
        mainWin.saveConfig()


class DialogFormulaParameters(Toplevel):
    def __init__(self, mainWin, options):
        parent = mainWin.parent
        self.modelManager = mainWin.modelManager
        super(DialogFormulaParameters, self).__init__(parent)
        self.parent = parent
        self.options = options
        parentGeometry = re.match(r"(\d+)x(\d+)[+]?([-]?\d+)[+]?([-]?\d+)", parent.geometry())
        dialogX = int(parentGeometry.group(3))
        dialogY = int(parentGeometry.group(4))
        self.accepted = False

        self.transient(self.parent)
        self.title(_("Formula Parameters and Trace Options"))

        frame = Frame(self)

        '''
        dialogFrame = Frame(frame, width=500)
        dialogFrame.columnconfigure(0, weight=1)
        dialogFrame.rowconfigure(0, weight=1)
        dialogFrame.grid(row=0, column=0, columnspan=4, sticky=(N, S, E, W), padx=3, pady=3)
        '''

        # mainWin.showStatus(_("loading formula options and parameters"))

        # load grid
        gridHdr(frame, 1, 0, "Parameters", columnspan=3)
        gridHdr(frame, 1, 1, "QName")
        gridHdr(frame, 2, 1, "Type")
        gridHdr(frame, 3, 1, "Value")

        self.gridCells = []
        y = 2
        dataTypes = ("xs:string", "xs:integer", "xs:decimal", "xs:boolean", "xs:date", "xs:datetime", "xs:QName")
        for parameter in options["parameterValues"].items():
            paramQname, paramTypeValue = parameter
            if isinstance(paramTypeValue, (tuple,list)):
                paramType, paramValue = paramTypeValue  # similar to modelTestcaseObject, where values() are (type,value)
            else:
                paramType = None
                paramValue = paramTypeValue
            self.gridCells.append( (
                gridCell(frame, 1, y, paramQname),
                gridCombobox(frame, 2, y, paramType, values=dataTypes),
                gridCell(frame, 3, y, paramValue)) )
            y += 1
        # extra entry for new cells
        for i in range(5):
            self.gridCells.append( (
                gridCell(frame, 1, y),
                gridCombobox(frame, 2, y, values=dataTypes),
                gridCell(frame, 3, y)) )
            y += 1
        y += 1

        # checkbox entries
        label(frame, 1, y, "Parameter Trace:")
        label(frame, 1, y + 3, "API Calls Trace:")
        label(frame, 1, y + 8, "Testcase Results & RSS Items:")
        label(frame, 2, y, "Variable Set Trace:")
        label(frame, 3, y, "Variables Trace:")
        self.checkboxes = (
           checkbox(frame, 1, y + 1,
                    "Expression Result",
                    "traceParameterExpressionResult"),
           checkbox(frame, 1, y + 2,
                    "Input Value",
                    "traceParameterInputValue"),
           checkbox(frame, 1, y + 4,
                    "Expression Source",
                    "traceCallExpressionSource"),
           checkbox(frame, 1, y + 5,
                    "Expression Code",
                    "traceCallExpressionCode"),
           checkbox(frame, 1, y + 6,
                    "Expression Evaluation",
                    "traceCallExpressionEvaluation"),
           checkbox(frame, 1, y + 7,
                    "Expression Result",
                    "traceCallExpressionResult"),
           checkbox(frame, 1, y + 9,
                    "Capture Warnings",
                    "testcaseResultsCaptureWarnings"),
           gridCombobox(frame, 1, y + 10, padx=24,
                        attr="testcaseResultOptions",
                        values=("match-any", "match-all")),

           checkbox(frame, 2, y + 1,
                    "Expression Source",
                    "traceVariableSetExpressionSource"),
           checkbox(frame, 2, y + 2,
                    "Expression Code",
                    "traceVariableSetExpressionCode"),
           checkbox(frame, 2, y + 3,
                    "Expression Evaluation",
                    "traceVariableSetExpressionEvaluation"),
           checkbox(frame, 2, y + 4,
                    "Expression Result",
                    "traceVariableSetExpressionResult"),
           checkbox(frame, 2, y + 5,
                    "Assertion Result Counts",
                    "traceAssertionResultCounts"),
           checkbox(frame, 2, y + 6,
                    "Assertion Satisfied [info]",
                    "traceSatisfiedAssertions"),
           checkbox(frame, 2, y + 7,
                    "Assertion Unsatisfied [error]",
                    "errorUnsatisfiedAssertions"),
           checkbox(frame, 2, y + 8,
                    "Assertion Unsatisfied [info]",
                    "traceUnsatisfiedAssertions"),
           checkbox(frame, 2, y + 9,
                    "Formula Rules",
                    "traceFormulaRules"),
           checkbox(frame, 2, y + 10,
                    "Evaluation Timing",
                    "timeVariableSetEvaluation"),
           checkbox(frame, 3, y + 1,
                    "Variable Dependencies",
                    "traceVariablesDependencies"),
           checkbox(frame, 3, y + 2,
                    "Variables Order",
                    "traceVariablesOrder"),
           checkbox(frame, 3, y + 3,
                    "Expression Source",
                    "traceVariableExpressionSource"),
           checkbox(frame, 3, y + 4,
                    "Expression Code",
                    "traceVariableExpressionCode"),
           checkbox(frame, 3, y + 5,
                    "Expression Evaluation",
                    "traceVariableExpressionEvaluation"),
           checkbox(frame, 3, y + 6,
                    "Expression Result",
                    "traceVariableExpressionResult"),
           checkbox(frame, 3, y + 7,
                    "Filter Winnowing",
                    "traceVariableFilterWinnowing"),
           checkbox(frame, 3, y + 8,
                    "Filters Result",
                    "traceVariableFiltersResult")

           # Note: if adding to this list keep ModelFormulaObject.FormulaOptions in sync

           )
        y += 11

        mainWin.showStatus(None)

        label(frame, 1, y, "IDs:")
        self.idsEntry = gridCell(frame, 1, y, options.get("runIDs"))
        self.idsEntry.grid(columnspan=2, padx=30)
        _w = 8 if self.modelManager.cntlr.isMac else 12
        okButton = Button(frame, text=_("OK"), width=_w, command=self.ok)
        cancelButton = Button(frame, text=_("Cancel"), width=_w, command=self.close)
        okButton.grid(row=y, column=3, sticky=W, pady=3)
        cancelButton.grid(row=y, column=3, sticky=E, pady=3, padx=3)

        frame.grid(row=0, column=0, sticky=(N,S,E,W))
        frame.columnconfigure(1, weight=3)
        frame.columnconfigure(2, weight=1)
        frame.columnconfigure(3, weight=3)
        window = self.winfo_toplevel()
        window.columnconfigure(0, weight=1)
        self.geometry("+{0}+{1}".format(dialogX+50,dialogY+100))

        #self.bind("<Return>", self.ok)
        #self.bind("<Escape>", self.close)

        self.protocol("WM_DELETE_WINDOW", self.close)
        self.grab_set()
        self.wait_window(self)

    def setOptions(self):
        # set formula options
        for checkbox in self.checkboxes:
            self.options[checkbox.attr] = checkbox.value
        parameterValues = {}
        for paramCells in self.gridCells:
            qnameCell, typeCell, valueCell = paramCells
            if qnameCell.value != "" and valueCell.value != "":
                # stored as strings, so they can be saved in json files
                parameterValues[qnameCell.value] = (typeCell.value, valueCell.value)
        self.options["parameterValues"] = parameterValues
        self.options["runIDs"] = self.idsEntry.value

    def ok(self, event=None):
        self.setOptions()
        self.accepted = True
        self.close()

    def close(self, event=None):
        self.parent.focus_set()
        self.destroy()



"""
See COPYRIGHT.md for copyright information.
"""
from __future__ import annotations

from collections import defaultdict
from typing import Any, cast

import regex as re

from arelle import XbrlConst
from arelle.ModelDtsObject import ModelConcept, ModelLink
from arelle.ModelObject import ModelObject
from arelle.PrototypeDtsObject import PrototypeObject
from arelle.ValidateXbrl import ValidateXbrl
from arelle.typing import TypeGetText
from .Util import isExtension, isInEsefTaxonomy

_: TypeGetText  # Handle gettext


def checkFilingDimensions(
    val: ValidateXbrl,
    defaultDimensionLinkroles: tuple[str, ...],
    lineItemsNotQualifiedLinkroles: tuple[str, ...],
) -> None:

    val.primaryItems = set() # concepts which are line items (should not also be dimension members
    val.domainMembers = set()  # concepts which are dimension domain members

    elrPrimaryItems = defaultdict(set)
    hcPrimaryItems: set[ModelConcept] = set()
    hcMembers: set[Any] = set()

    def addDomMbrs(sourceDomMbr: ModelConcept, ELR: str, membersSet: set[ModelConcept]) -> None:
        if isinstance(sourceDomMbr, ModelConcept) and sourceDomMbr not in membersSet:
            membersSet.add(sourceDomMbr)
            for domMbrRel in val.modelXbrl.relationshipSet(XbrlConst.domainMember, ELR).fromModelObject(sourceDomMbr):
                addDomMbrs(domMbrRel.toModelObject, domMbrRel.consecutiveLinkrole, membersSet)

    for hasHypercubeArcrole in (XbrlConst.all, XbrlConst.notAll):
        hasHypercubeRelationships = val.modelXbrl.relationshipSet(hasHypercubeArcrole).fromModelObjects()

        for hasHcRels in hasHypercubeRelationships.values():
            for hasHcRel in hasHcRels:
                sourceConcept: ModelConcept = hasHcRel.fromModelObject
                hcPrimaryItems.add(sourceConcept)
                # find associated primary items to source concept
                for domMbrRel in val.modelXbrl.relationshipSet(XbrlConst.domainMember).fromModelObject(sourceConcept):
                    if domMbrRel.consecutiveLinkrole == hasHcRel.linkrole: # only those related to this hc
                        addDomMbrs(domMbrRel.toModelObject, domMbrRel.consecutiveLinkrole, hcPrimaryItems)
                val.primaryItems.update(hcPrimaryItems)
                hc = hasHcRel.toModelObject
                if hasHypercubeArcrole == XbrlConst.all:
                    if not hasHcRel.isClosed and isExtension(val, hasHcRel):
                        val.modelXbrl.error("ESEF.3.4.2.openPositiveHypercubeInDefinitionLinkbase",
                            _("Hypercubes appearing as target of definition arc with http://xbrl.org/int/dim/arcrole/all arcrole MUST have xbrldt:closed attribute set to \"true\""
                              ": hypercube %(hypercube)s, linkrole %(linkrole)s, primary item %(primaryItem)s"),
                            modelObject=hasHcRel, hypercube=hc.qname, linkrole=hasHcRel.linkrole, primaryItem=sourceConcept.qname)
                elif hasHypercubeArcrole == XbrlConst.notAll:
                    if hasHcRel.isClosed and isExtension(val, hasHcRel):
                        val.modelXbrl.error("ESEF.3.4.2.closedNegativeHypercubeInDefinitionLinkbase",
                            _("Hypercubes appearing as target of definition arc with http://xbrl.org/int/dim/arcrole/notAll arcrole MUST have xbrldt:closed attribute set to \"false\""
                              ": hypercube %(hypercube)s, linkrole %(linkrole)s, primary item %(primaryItem)s"),
                            modelObject=hasHcRel, hypercube=hc.qname, linkrole=hasHcRel.linkrole, primaryItem=sourceConcept.qname)
                    if isExtension(val, hasHcRel):
                        val.modelXbrl.error("ESEF.3.4.2.notAllArcroleUsedInDefinitionLinkbase",
                            _("Extension taxonomies MUST NOT define definition arcs with http://xbrl.org/int/dim/arcrole/notAll arcrole"
                              ": hypercube %(hypercube)s, linkrole %(linkrole)s, primary item %(primaryItem)s"),
                            modelObject=hasHcRel, hypercube=hc.qname, linkrole=hasHcRel.linkrole, primaryItem=sourceConcept.qname)
                for hcDimRel in val.modelXbrl.relationshipSet(XbrlConst.hypercubeDimension, hasHcRel.consecutiveLinkrole).fromModelObject(hc):
                    dim = hcDimRel.toModelObject
                    if isinstance(dim, ModelConcept):
                        for dimDomRel in val.modelXbrl.relationshipSet(XbrlConst.dimensionDomain, hcDimRel.consecutiveLinkrole).fromModelObject(dim):
                            dom = dimDomRel.toModelObject
                            if isinstance(dom, ModelConcept):
                                 addDomMbrs(dom, dimDomRel.consecutiveLinkrole, hcMembers)
                val.domainMembers.update(hcMembers)
                if hasHcRel.linkrole in lineItemsNotQualifiedLinkroles or hcMembers:
                    for hcPrimaryItem in hcPrimaryItems:
                        if not hcPrimaryItem.isAbstract:
                            elrPrimaryItems[hasHcRel.linkrole].add(hcPrimaryItem)
                            elrPrimaryItems["*"].add(hcPrimaryItem) # members of any ELR
                hcPrimaryItems.clear()
                hcMembers.clear()

    # reported pri items not in LineItemsNotQualifiedLinkrole
    nsExcl = val.authParam.get("lineItemsNotDimQualExclusionNsPattern")
    if nsExcl:
        nsExclPat = re.compile(nsExcl)
    i = set(concept
            for qn, facts in val.modelXbrl.factsByQname.items()
            if any(not f.context.qnameDims for f in facts if f.context is not None)
            for concept in (val.modelXbrl.qnameConcepts.get(qn),)
            if concept is not None and
               not any(concept in elrPrimaryItems.get(lr, set()) for lr in lineItemsNotQualifiedLinkroles) and
               concept not in elrPrimaryItems.get("*", set()) and
               (not nsExcl or not nsExclPat.match(cast(str, qn.namespaceURI))))
    if i:
        val.modelXbrl.error("ESEF.3.4.2.extensionTaxonomyLineItemNotLinkedToAnyHypercube",
            _("Line items that do not require any dimensional information to tag data MUST be linked to the dedicated \"Line items not dimensionally qualified\" hypercube in %(linkrole)s declared in esef_cor.xsd, primary item %(qnames)s"),
            modelObject=i, linkrole=lineItemsNotQualifiedLinkroles[0], qnames=", ".join(sorted(str(c.qname) for c in i)))

    # check ELRs with WiderNarrower relationships
    elrsContainingDimensionalRelationships = set(
        ELR
        for arcrole, ELR, linkqname, arcqname in val.modelXbrl.baseSets.keys()
        if arcrole == "XBRL-dimensions" and ELR is not None)
    anchorsInDimensionalELRs = defaultdict(list)
    for anchoringRel in val.modelXbrl.relationshipSet(XbrlConst.widerNarrower).modelRelationships:
        ELR = anchoringRel.linkrole
        fromObj = anchoringRel.fromModelObject
        toObj = anchoringRel.toModelObject
        if fromObj is not None and toObj is not None and fromObj.type is not None and toObj.type is not None:
            if not (isInEsefTaxonomy(val, fromObj) ^ isInEsefTaxonomy(val, toObj)):
                val.modelXbrl.error("ESEF.3.3.1.anchoringRelationshipBase",
                    _("Anchoring relationships MUST be from or to an ESEF element, from %(qname1)s to %(qname2)s"),
                    modelObject=(anchoringRel, fromObj, toObj), qname1=fromObj.qname, qname2=toObj.qname)
            if fromObj.type.isDomainItemType or toObj.type.isDomainItemType:
                val.modelXbrl.error("ESEF.3.3.1.anchoringRelationshipsForDomainMembersDefinedUsingWiderNarrowerArcrole",
                    _("Anchoring relationships MUST be from and to concepts, from %(qname1)s to %(qname2)s"),
                    modelObject=(anchoringRel, fromObj, toObj), qname1=fromObj.qname, qname2=toObj.qname)
            elif fromObj.isDimensionItem or toObj.isDimensionItem:
                val.modelXbrl.error("ESEF.3.3.1.anchoringRelationshipsForDimensionsDefinedUsingWiderNarrowerArcrole",
                    _("Anchoring relationships MUST be from and to concepts, from %(qname1)s to %(qname2)s"),
                    modelObject=(anchoringRel, fromObj, toObj), qname1=fromObj.qname, qname2=toObj.qname)
            else: # neither from nor to are dimensions or domain members
                if ELR in elrsContainingDimensionalRelationships:
                    anchorsInDimensionalELRs[ELR].append(anchoringRel)
    if anchorsInDimensionalELRs:
        for ELR, rels in anchorsInDimensionalELRs.items():
            val.modelXbrl.error("ESEF.3.3.2.anchoringRelationshipsForConceptsDefinedInElrContainingDimensionalRelationships",
                _("Anchoring relationships for concepts MUST be defined in a dedicated extended link role (or roles if needed to properly represent the relationships), e.g. http://{issuer default pattern for roles}/Anchoring. %(anchoringDimensionalELR)s"),
                modelObject=rels, anchoringDimensionalELR=ELR)

    # check base set dimension default overrides in extension taxonomies
    for modelLink in cast(list[ModelLink], val.modelXbrl.baseSets[XbrlConst.dimensionDefault, None, None, None]):
        if isExtension(val, modelLink):
            for linkChild in modelLink:
                if (isinstance(linkChild,(ModelObject,PrototypeObject)) and
                    linkChild.get("{http://www.w3.org/1999/xlink}type") == "arc" and
                    linkChild.get("{http://www.w3.org/1999/xlink}arcrole") == XbrlConst.dimensionDefault):
                    fromLabel = linkChild.get("{http://www.w3.org/1999/xlink}from")
                    if fromLabel is None:
                        continue
                    for fromResource in modelLink.labeledResources[fromLabel]:
                        if not isExtension(val, fromResource):
                            val.modelXbrl.error("ESEF.3.4.3.extensionTaxonomyOverridesDefaultMembers",
                                _("The extension taxonomy MUST not modify (prohibit and/or override) default members assigned to dimensions by the ESEF taxonomy."),
                                modelObject=linkChild)
                    if modelLink.role not in defaultDimensionLinkroles:
                        val.modelXbrl.error("ESEF.3.4.3.dimensionDefaultLinkrole",
                            _("Each dimension in an issuer specific extension taxonomy MUST be assigned to a default member in the ELR with role URI http://www.esma.europa.eu/xbrl/role/cor/ifrs-dim_role-990000, but linkrole used is %(linkrole)s."),
                            modelObject=linkChild, linkrole=modelLink.role)



'''
See COPYRIGHT.md for copyright information.
'''
from tkinter import *
try:
    from tkinter.ttk import *
except ImportError:
    from ttk import *
import os
from arelle import ViewWinTree
from arelle.ModelDocumentType import ModelDocumentType

def viewDTS(modelXbrl, tabWin, altTabWin=None):
    view = ViewDTS(modelXbrl, tabWin)
    modelXbrl.modelManager.showStatus(_("viewing DTS"))
    view.view()

    menu = view.contextMenu()
    view.menuAddExpandCollapse()
    view.menuAddClipboard()
    view.menuAddViews(addClose=False, tabWin=altTabWin)

class ViewDTS(ViewWinTree.ViewTree):
    def __init__(self, modelXbrl, tabWin):
        super(ViewDTS, self).__init__(modelXbrl, tabWin, "DTS", True)

    def view(self):
        self.clearTreeView()
        self.viewDtsElement(self.modelXbrl.modelDocument, "", 1, set(), {self.modelXbrl.modelDocument})


    def viewDtsElement(self, modelDocument, parentNode, n, parents, siblings):
        if modelDocument.type == ModelDocumentType.INLINEXBRLDOCUMENTSET:
            if modelDocument.entrypoint is not None and "id" in modelDocument.entrypoint:
                text = f"{modelDocument.entrypoint['id']} (IXDS)"
            else:
                text = "Inline XBRL Document Set" # no file name or ID to display
        else:
            text = "{0} - {1}".format(os.path.basename(modelDocument.uri), modelDocument.gettype())
        node = self.treeView.insert(parentNode, "end",
                    text=text,
                    tags=("odd" if n & 1 else "even",))
        children = modelDocument.referencesDocument.keys()
        childFamily = parents | siblings
        for i, referencedDocument in enumerate(sorted(children, key=lambda d: d.objectIndex)): # provide consistent order
            if referencedDocument not in parents:
                self.viewDtsElement(referencedDocument, node, n + i + 1, childFamily, children)

    def viewModelObject(self, modelObject):
        pass



# -*- coding: utf-8 -*-
'''
saveOIMTaxonomy.py is a plug-in that saves an extension taxonomy in the json OIM taxonomy format

See COPYRIGHT.md for copyright information.
'''
import os, io, json
from arelle.Version import authorLabel, copyrightLabel
from arelle import XbrlConst
from collections import OrderedDict

jsonDocumentType = "https://xbrl.org/PWD/2023-05-17/cti"
jsonTxmyVersion = "1.0"
primaryLang = "en"

excludeImportNamespaces = {XbrlConst.xbrli, XbrlConst.xbrldt}

def saveOIMTaxonomy(dts, jsonFile):
    from arelle.ModelDocumentType import ModelDocumentType

    # identify extension schema
    namespacePrefixes = dict((ns, prefix) for prefix, ns in dts.prefixedNamespaces.items())
    namespacesInUse = set()
    extensionSchemaDoc = None
    if dts.modelDocument.type == ModelDocumentType.SCHEMA:
        extensionSchemaDoc = dts.modelDocument
    elif dts.modelDocument.type in (ModelDocumentType.INSTANCE, ModelDocumentType.INLINEXBRL, ModelDocumentType.INLINEXBRLDOCUMENTSET):
        for doc, docReference in dts.modelDocument.referencesDocument.items():
            if "href" in docReference.referenceTypes:
                extensionSchemaDoc = doc
                extensionPrefix = namespacePrefixes[extensionSchemaDoc.targetNamespace]
                namespacesInUse.add(extensionSchemaDoc.targetNamespace)
                break
    if extensionSchemaDoc is None:
        dts.info("error:saveLoadableOIMTaxonomy",
         _("Unable to identify extension taxonomy."),
         modelObject=dts)
        return

    oimTxmy = OrderedDict()
    oimTxmy["documentInfo"] = docInfo = OrderedDict()
    docInfo["documentType"] = jsonDocumentType
    docInfo["namespaces"] = namespaces = []
    oimTxmy["taxonomy"] = txmy = OrderedDict()

    # provide consistent order to taxonomy properties and objects
    txmy["name"] = os.path.splitext(extensionSchemaDoc.basename)[0]
    txmy["namespace"] = extensionSchemaDoc.targetNamespace
    txmy["version"] = jsonTxmyVersion
    txmy["entryPoint"] = os.path.basename(jsonFile)
    txmy["importedTaxonomies"] = imports = []
    txmy["concepts"] = concepts = []
    txmy["cubes"] = cubes = []
    txmy["domains"] = domains = []
    txmy["networks"] = networks = []
    txmy["relationships"] = relationships = []
    txmy["labels"] = labels = []
    sharedLabelRefs = {}
    txmy["references"] = references = []

    # taxonomy object
    extensionLinkbaseDocs = {doc
                             for doc, docReference in extensionSchemaDoc.referencesDocument.items()
                             if "href" in docReference.referenceTypes and docReference.referringModelObject.elementQname == XbrlConst.qnLinkLinkbaseRef}
    extensionLinkbaseDocs.add(extensionSchemaDoc)

    labelsRelationshipSet = dts.relationshipSet(XbrlConst.conceptLabel)
    for concept in sorted(set(dts.qnameConcepts.values()), key=lambda c:c.name): # may be twice if unqualified, with and without namespace
        if concept.modelDocument == extensionSchemaDoc:
            c = OrderedDict()
            concepts.append(c)
            c["name"] = str(concept.qname)
            if concept.typeQname: # may be absent
                c["dataType"] = str(concept.typeQname)
            if concept.substitutionGroupQname:
                c["substitutionGroup"] = str(concept.substitutionGroupQname)
            if concept.periodType:
                c["periodType"] = concept.periodType
            if concept.balance:
                c["balance"] = concept.balance
            c["abstract"] = str(concept.abstract).lower()
            c["nillable"] = str(concept.nillable).lower()
        conceptLabels = dict(((lbl.role, lbl.xmlLang), lbl)
                             for rel in labelsRelationshipSet.fromModelObject(concept)
                             for lbl in (rel.toModelObject,)
                             if lbl is not None)
        for _key, label in sorted(conceptLabels.items(), key=lambda i:i[0]):
            labelDupKey = (label.xmlLang, label.role, label.textValue)
            if labelDupKey in sharedLabelRefs:
                sharedLabelRefs[labelDupKey]["relatedId"].append(str(concept.qname))
            else:
                l = OrderedDict((
                    ("relatedId", [str(concept.qname)]),
                    ("language", label.xmlLang),
                    ("labelType", os.path.basename(label.role)),
                    ("value", label.textValue)))
                labels.append(l)
                sharedLabelRefs[labelDupKey] = l
        # TBD add references here

    # extended link roles defined in this document
    networkOrder = 0
    for roleURI, roleTypes in sorted(dts.roleTypes.items(),
            # sort on definition if any else URI
            key=lambda item: (item[1][0].definition if len(item[1]) and item[1][0].definition else item[0])):
        for roleType in roleTypes:
            if roleType.modelDocument == extensionSchemaDoc:
                name = f"{extensionPrefix}:_{os.path.basename(roleURI)}_"
                definition = roleType.definition
                # define network concept (not in XBRL 2.1
                c = OrderedDict((("name", name),
                                 ("dataType","dtr:networkItemType"),
                                 ("periodType", "duration"),
                                 ("substitutionGroup", "xbrli:item"),
                                 ("abstract", True),
                                 ("nillable", True)))
                concepts.append(c)
                networks.append( OrderedDict((
                    ("networkURI", roleType.roleURI),
                    ("name", name))))
                networkOrder += 1
                relationships.append( OrderedDict((
                    ("source", "xbrli:networkConcept"),
                    ("target", name),
                    ("order", networkOrder),
                    ("networkURI", "https://www.xbrl.org/Network"),
                    ("relationshipType", "root-child"))))
                if definition:
                    labels.append( OrderedDict((
                        ("relatedId", [name]),
                        ("language", primaryLang),
                        ("labelType", XbrlConst.standardLabel),
                        ("value", definition))))

    # cubes by linkrole
    for arcrole, linkrole, arcQN, linkQN in dts.baseSets.keys():
        if arcrole == XbrlConst.all and linkrole is not None and arcQN is None and linkQN is None:
            for rel in dts.relationshipSet(XbrlConst.all, linkrole).modelRelationships:
                if rel.modelDocument not in extensionLinkbaseDocs:
                    continue
                hc = rel.toModelObject
                priItem = rel.fromModelObject
                if hc is not None and priItem is not None:
                    dims = []
                    dim = OrderedDict((("dimensionConcept", "xbrl:PrimaryDimension"),
                                       ("domainID", priItem.name),
                                       ("dimensionType", "xbrl:concept")))
                    dims.append(dim)
                    # priItem domains
                    rels = []
                    domains.append( OrderedDict((
                        ("domainId", priItem.name),
                        ("domainConcept", str(priItem.qname)),
                        ("networkURI", rel.linkrole),
                        ("relationships", rels))) )
                    for domRel in dts.relationshipSet(XbrlConst.domainMember, rel.consecutiveLinkrole).fromModelObject(priItem):
                        domObj = domRel.toModelObject
                        if domObj is not None:
                            rels.append( OrderedDict((
                                ("source", str(priItem.qname)),
                                ("target", str(domObj.qname)),
                                ("order", domRel.order))) )
                    # dimension domains
                    for dimRel in dts.relationshipSet(XbrlConst.hypercubeDimension, rel.consecutiveLinkrole).fromModelObject(hc):
                        dimObj = dimRel.toModelObject
                        if dimObj is not None:
                            dim = OrderedDict((("dimensionConcept", str(dimObj.qname)),
                                               ("domainId", dimObj.name),
                                               ("dimensionType", "explicit")))
                            dims.append(dim)
                            domRels = []
                            domains.append( OrderedDict((
                                ("domainId", dimObj.name),
                                ("domainConcept", str(dimObj.qname)),
                                ("networkURI", rel.linkrole),
                                ("relationships", domRels))) )
                            for domRel in dts.relationshipSet(XbrlConst.dimensionDomain, rel.consecutiveLinkrole).fromModelObject(dimObj):
                                domObj = domRel.toModelObject
                                if domObj is not None:
                                    domRels.append( OrderedDict((
                                        ("source", str(dimObj.qname)),
                                        ("target", str(domObj.qname)),
                                        ("order", domRel.order))) )
                cubes.append( OrderedDict((
                    ("name", str(hc.qname)),
                    ("networkURI", rel.linkrole),
                    ("cubeType", "aggregate"),
                    ("dimensions", dims))) )

    # domains

    # tree walk recursive function
    def treeWalk(depth, concept, arcrole, relSet, visited):
        if concept is not None:
            if concept not in visited:
                visited.add(concept)
                for rel in relSet.fromModelObject(concept):
                    toConcept = rel.toModelObject
                    if toConcept is not None:
                        namespacesInUse.add(concept.qname.namespaceURI)
                        namespacesInUse.add(toConcept.qname.namespaceURI)
                        if rel.modelDocument in extensionLinkbaseDocs:
                            relationships.append( OrderedDict((
                                ("source", str(concept.qname)),
                                ("target", str(toConcept.qname)),
                                ("order", rel.order),
                                ("networkURI", rel.linkrole),
                                ("relationshipType", os.path.basename(rel.arcrole))) +
                                ((("weight", rel.weight),) if rel.weight is not None else ())
                                ) )
                        row = treeWalk(depth + 1, toConcept, arcrole, relSet, visited)
                visited.remove(concept)

    # use presentation relationships for conceptsWs
    for arcrole in (XbrlConst.parentChild,) + XbrlConst.summationItems:
        # sort URIs by definition
        linkroleUris = []
        relationshipSet = dts.relationshipSet(arcrole)
        if relationshipSet:
            for linkroleUri in relationshipSet.linkRoleUris:
                modelRoleTypes = dts.roleTypes.get(linkroleUri)
                if modelRoleTypes:
                    roledefinition = (modelRoleTypes[0].genLabel(strip=True) or modelRoleTypes[0].definition or linkroleUri)
                else:
                    roledefinition = linkroleUri
                linkroleUris.append((roledefinition, linkroleUri))
            linkroleUris.sort()

            # for each URI in definition order
            for roledefinition, linkroleUri in linkroleUris:
                # elr relationships for tree walk
                linkRelationshipSet = dts.relationshipSet(arcrole, linkroleUri)
                for rootConcept in linkRelationshipSet.rootConcepts:
                    # is root child required?
                    treeWalk(0, rootConcept, arcrole, linkRelationshipSet, set())

    for ns in sorted(namespacesInUse, key=lambda ns: namespacePrefixes[ns]):
        namespaces.append(OrderedDict((("prefix", namespacePrefixes[ns]),
                                       ("uri", ns))))

    try:
        with open(jsonFile, "w") as fh:
            fh.write(json.dumps(oimTxmy, indent=3))

        dts.info("info:saveOIMTaxonomy",
            _("Saved OIM Taxonomy file: %(excelFile)s"),
            excelFile=os.path.basename(jsonFile),
            modelXbrl=dts)
    except Exception as ex:
        dts.error("exception:saveOIMTaxonomy",
            _("File saving exception: %(error)s"), error=ex,
            modelXbrl=dts)

def saveOIMTaxonomyMenuEntender(cntlr, menu, *args, **kwargs):
    # Extend menu with an item for the savedts plugin
    menu.add_command(label="Save OIM Taxonomy",
                     underline=0,
                     command=lambda: saveOIMTaxonomyMenuCommand(cntlr) )

def saveOIMTaxonomyMenuCommand(cntlr):
    # save DTS menu item has been invoked
    if cntlr.modelManager is None or cntlr.modelManager.modelXbrl is None:
        cntlr.addToLog("No taxonomy loaded.")
        return
        # get file name into which to save log file while in foreground thread
    jsonFile = cntlr.uiFileDialog("save",
            title=_("arelle - Save OIM Taxonomy file"),
            initialdir=cntlr.config.setdefault("OIMTaxonomyFileDir","."),
            filetypes=[(_("OIM Taxonomy .json"), "*.json")],
            defaultextension=".json")
    if not jsonFile:
        return False
    import os
    cntlr.config["OIMTaxonomyFileDir"] = os.path.dirname(jsonFile)
    cntlr.saveConfig()

    import threading
    thread = threading.Thread(target=lambda
                                  _dts=cntlr.modelManager.modelXbrl,
                                  _jsonFile=jsonFile:
                                        saveOIMTaxonomy(_dts, _jsonFile))
    thread.daemon = True
    thread.start()

def saveOIMTaxonomyCommandLineOptionExtender(parser, *args, **kwargs):
    # extend command line options with a save DTS option
    parser.add_option("--save-OIM-taxonomy",
                      dest="saveOIMTaxonomy",
                      help=_("Save OIM Taxonomy file"))

def saveOIMTaxonomyCommandLineXbrlRun(cntlr, options, modelXbrl, *args, **kwargs):
    # extend XBRL-loaded run processing for this option
    jsonFile = getattr(options, "saveOIMTaxonomy", None)
    if jsonFile:
        if cntlr.modelManager is None or cntlr.modelManager.modelXbrl is None:
            cntlr.addToLog("No taxonomy loaded.")
            return
        saveOIMTaxonomy(cntlr.modelManager.modelXbrl, jsonFile)

__pluginInfo__ = {
    'name': 'Save OIM Taxonomy',
    'version': '0.9',
    'description': "This plug-in saves an OIM Taxonomy.",
    'license': 'Apache-2',
    'author': authorLabel,
    'copyright': copyrightLabel,
    # classes of mount points (required)
    'CntlrWinMain.Menu.Tools': saveOIMTaxonomyMenuEntender,
    'CntlrCmdLine.Options': saveOIMTaxonomyCommandLineOptionExtender,
    'CntlrCmdLine.Xbrl.Run': saveOIMTaxonomyCommandLineXbrlRun,
}



"""
See COPYRIGHT.md for copyright information.
"""
from __future__ import annotations

import logging
import os
from collections import defaultdict
from decimal import Decimal
from typing import TYPE_CHECKING, Any, Union, cast

from arelle import UrlUtil, XmlUtil, ModelValue, XbrlConst
from arelle.FileSource import FileSource
from arelle.Locale import format_string
from arelle.ModelObject import ModelObject, ObjectPropertyViewWrapper
from arelle.PluginManager import hasPluginWithHook, pluginClassMethods
from arelle.PythonUtil import flattenSequence

if TYPE_CHECKING:
    from arelle.ModelManager import ModelManager
    from arelle.ModelXbrl import ModelXbrl
    from arelle.typing import EmptyTuple

LoggableValue = Union[str, dict[Any, Any], list[Any], set[Any], tuple[Any, ...]]
EMPTY_TUPLE: EmptyTuple = ()


class ErrorManager:
    logHasRelevelerPlugin: bool | None
    _errorCaptureLevel: int
    _errors: list[str | None]
    _logCount: dict[str, int] = {}
    _logRefFileRelUris: defaultdict[Any, dict[str, str]]
    _modelManager: ModelManager

    def __init__(self, modelManager: ModelManager, errorCaptureLevel: int):
        self._errorCaptureLevel = errorCaptureLevel
        self._errors = []
        self._logCount = {}
        self._logRefFileRelUris = defaultdict(dict)
        self._modelManager = modelManager
        self.logHasRelevelerPlugin = None

    @property
    def errors(self) -> list[str | None]:
        return self._errors

    @property
    def logCount(self) -> dict[str, int]:
        return self._logCount

    def _effectiveMessageCode(self, messageCodes: tuple[Any] | str) -> str | None:
        """
        If codes includes EFM, GFM, HMRC, or SBR-coded error then the code chosen (if a sequence)
        corresponds to whether EFM, GFM, HMRC, or SBR validation is in effect.
        """
        effectiveMessageCode = None
        _validationType = self._modelManager.disclosureSystem.validationType
        _exclusiveTypesPattern = self._modelManager.disclosureSystem.exclusiveTypesPattern

        for argCode in messageCodes if isinstance(messageCodes,tuple) else (messageCodes,):
            if (isinstance(argCode, ModelValue.QName) or
                    (_validationType and argCode and argCode.startswith(_validationType)) or
                    (not _exclusiveTypesPattern or _exclusiveTypesPattern.match(argCode or "") == None)):
                effectiveMessageCode = argCode
                break
        return effectiveMessageCode

    def clear(self) -> None:
        self._errors.clear()
        self._logCount.clear()

    def isLoggingEffectiveFor(self, logger: logging.Logger, **kwargs: Any) -> bool:  # args can be messageCode(s) and level
        messageCodeFilter = getattr(logger, 'messageCodeFilter', None)
        messageLevelFilter = getattr(logger, 'messageLevelFilter', None)
        if "messageCodes" in kwargs or "messageCode" in kwargs:
            if "messageCodes" in kwargs:
                messageCodes = kwargs["messageCodes"]
            else:
                messageCodes = kwargs["messageCode"]
            messageCode = self._effectiveMessageCode(messageCodes)
            codeEffective = (messageCode and
                             (not messageCodeFilter or messageCodeFilter.match(messageCode)))
        else:
            codeEffective = True
        if "level" in kwargs and messageLevelFilter:
            levelEffective = messageLevelFilter.match(kwargs["level"].lower())
        else:
            levelEffective = True
        return bool(codeEffective and levelEffective)

    def log(
            self,
            logger: logging.Logger,
            level: str,
            codes: Any,
            msg: str,
            sourceModelXbrl: ModelXbrl | None = None,
            fileSource: FileSource | None = None,
            entryLoadingUrl: str | None = None,
            logRefObjectProperties: bool = False,
            **args: Any
    ) -> None:
        """Same as error(), but level passed in as argument
        """
        messageCodeFilter = getattr(logger, 'messageCodeFilter', None)
        messageLevelFilter = getattr(logger, 'messageLevelFilter', None)
        # determine logCode
        messageCode = self._effectiveMessageCode(codes)
        if messageCode == "asrtNoLog":
            self._errors.append(args["assertionResults"])
            return
        if self.logHasRelevelerPlugin is None:
            self.logHasRelevelerPlugin = hasPluginWithHook("Logging.Severity.Releveler")
        if sourceModelXbrl is not None and self.logHasRelevelerPlugin:
            for pluginXbrlMethod in pluginClassMethods("Logging.Severity.Releveler"):
                level, messageCode = pluginXbrlMethod(sourceModelXbrl, level, messageCode, args) # args must be passed as dict because it may contain modelXbrl or messageCode key value
        if (messageCode and
                (not messageCodeFilter or messageCodeFilter.match(messageCode)) and
                (not messageLevelFilter or messageLevelFilter.match(level.lower()))):
            # note that plugin Logging.Message.Parameters may rewrite messageCode which now occurs after filtering on messageCode
            messageCode, logArgs, extras = self._logArguments(
                messageCode,
                msg,
                args,
                sourceModelXbrl=sourceModelXbrl,
                fileSource=fileSource,
                entryLoadingUrl=entryLoadingUrl,
                logRefObjectProperties=logRefObjectProperties,
            )
            numericLevel = logging._checkLevel(level)  #type: ignore[attr-defined]
            self._logCount[numericLevel] = self._logCount.get(numericLevel, 0) + 1
            if numericLevel >= self._errorCaptureLevel:
                try: # if there's a numeric errorCount arg, extend messages codes by count
                    self._errors.extend([messageCode] * int(logArgs[1]["errorCount"]))
                except (IndexError, KeyError, ValueError): # no msgArgs, no errorCount, or not int
                    self._errors.append(messageCode) # assume one error occurence
            """@messageCatalog=[]"""
            logger.log(numericLevel, *logArgs, exc_info=args.get("exc_info"), extra=extras)

    def _logArguments(
            self,
            messageCode: str,
            msg: str,
            codedArgs: dict[str, str],
            sourceModelXbrl: ModelXbrl | None = None,
            fileSource: FileSource | None = None,
            entryLoadingUrl: str | None = None,
            logRefObjectProperties: bool = False,
    ) -> Any:
        # Prepares arguments for logger function as per info() below.

        def propValues(properties: Any) -> Any:
            # deref objects in properties
            return [(p[0], str(p[1])) if len(p) == 2 else (p[0], str(p[1]), propValues(p[2]))
                    for p in properties if 2 <= len(p) <= 3]
        # determine message and extra arguments
        fmtArgs: dict[str, LoggableValue] = {}
        extras: dict[str, Any] = {"messageCode":messageCode}
        modelObjectArgs: tuple[Any, ...] | list[Any] = ()
        sourceModelDocument = getattr(sourceModelXbrl, "modelDocument", None)
        for argName, argValue in codedArgs.items():
            if argName in ("modelObject", "modelXbrl", "modelDocument"):
                if sourceModelDocument is not None:
                    entryUrl = sourceModelDocument.uri
                else:
                    if entryLoadingUrl is not None:
                        entryUrl = entryLoadingUrl
                    else:
                        assert fileSource is not None, 'Expected FileSource to be available for fallback entry URL.'
                        entryUrl = fileSource.url
                refs: list[dict[str, Any]] = []
                modelObjectArgs_complex = argValue if isinstance(argValue, (tuple,list,set)) else (argValue,)
                modelObjectArgs = flattenSequence(modelObjectArgs_complex)
                for arg in modelObjectArgs:
                    if arg is not None:
                        if isinstance(arg, str):
                            objectUrl = arg
                        else:
                            try:
                                if isinstance(arg, ObjectPropertyViewWrapper):
                                    objectUrl = arg.modelObject.modelDocument.displayUri
                                else:
                                    objectUrl = arg.modelDocument.displayUri
                            except AttributeError:
                                try:
                                    objectUrl = arg.displayUri
                                except AttributeError:
                                    try:
                                        objectUrl = sourceModelDocument.displayUri  # type: ignore[union-attr]
                                    except AttributeError:
                                        objectUrl = entryLoadingUrl or ""
                        try:
                            if objectUrl.endswith("/_IXDS"):
                                file = objectUrl[:-6] # inline document set or report package
                            elif objectUrl in self._logRefFileRelUris.get(entryUrl, EMPTY_TUPLE):
                                file = self._logRefFileRelUris[entryUrl][objectUrl]
                            else:
                                file = UrlUtil.relativeUri(entryUrl, objectUrl)
                                self._logRefFileRelUris[entryUrl][objectUrl] = file
                        except:
                            file = ""
                        ref: dict[str, Any] = {}
                        if isinstance(arg,(ModelObject, ObjectPropertyViewWrapper)):
                            _arg:ModelObject = arg.modelObject if isinstance(arg, ObjectPropertyViewWrapper) else arg
                            if len(modelObjectArgs) > 1 and getattr(arg,"tag",None) == "instance":
                                continue # skip IXDS top level element
                            fragmentIdentifier = "#" + cast(str, XmlUtil.elementFragmentIdentifier(_arg))
                            if not hasattr(_arg, 'modelDocument') and _arg.namespaceURI == XbrlConst.svg and len(refs) > 0:
                                # This is an embedded SVG document without its own file.
                                # Set the href to the containing document element that defined the encoded SVG.
                                # and define a nestedHrefs attribute with the fragment identifier.
                                priorRef = refs[-1]
                                ref["href"] = priorRef["href"]
                                priorNestedHrefs = priorRef.get("customAttributes", {}).get("nestedHrefs", [])
                                ref["customAttributes"] = {
                                    "nestedHrefs": [*priorNestedHrefs, fragmentIdentifier]
                                }
                                if priorArgSourceline := priorRef.get("sourceLine"):
                                    ref["sourceLine"] = priorArgSourceline
                            else:
                                ref["href"] = file + fragmentIdentifier
                                ref["sourceLine"] = _arg.sourceline
                            ref["objectId"] = _arg.objectId()
                            if logRefObjectProperties:
                                try:
                                    ref["properties"] = propValues(arg.propertyView)
                                except AttributeError:
                                    pass # is a default properties entry appropriate or needed?
                            if any(True for m in pluginClassMethods("Logging.Ref.Properties")):
                                refProperties: Any = ref.get("properties", {})
                                for pluginXbrlMethod in pluginClassMethods("Logging.Ref.Properties"):
                                    pluginXbrlMethod(arg, refProperties, codedArgs)
                                if refProperties:
                                    ref["properties"] = refProperties
                        else:
                            ref["href"] = file
                            try:
                                ref["sourceLine"] = arg.sourceline
                            except AttributeError:
                                pass # arg may not have sourceline, ignore if so
                        if any(True for m in pluginClassMethods("Logging.Ref.Attributes")):
                            refAttributes: dict[str, str] = {}
                            for pluginXbrlMethod in pluginClassMethods("Logging.Ref.Attributes"):
                                pluginXbrlMethod(arg, refAttributes, codedArgs)
                            if refAttributes:
                                ref["customAttributes"] = refAttributes
                        refs.append(ref)
                extras["refs"] = refs
            elif argName == "sourceFileLine":
                # sourceFileLines is pairs of file and line numbers, e.g., ((file,line),(file2,line2),...)
                ref = {}
                if isinstance(argValue, (tuple,list)):
                    ref["href"] = str(argValue[0])
                    if len(argValue) > 1 and argValue[1]:
                        ref["sourceLine"] = str(argValue[1])
                else:
                    ref["href"] = str(argValue)
                extras["refs"] = [ref]
            elif argName == "sourceFileLines":
                # sourceFileLines is tuple/list of pairs of file and line numbers, e.g., ((file,line),(file2,line2),...)
                sf_refs: list[dict[str, str]] = []
                argvalues: tuple[Any, ...] | list[Any] = argValue if isinstance(argValue, (tuple, list)) else (argValue,)
                for arg in argvalues:
                    ref = {}
                    if isinstance(arg, (tuple, list)):
                        arg_: tuple[Any, ...] | list[Any] = arg
                        ref["href"] = str(arg_[0])
                        if len(arg_) > 1 and arg_[1]:
                            ref["sourceLine"] = str(arg_[1])
                    else:
                        ref["href"] = str(arg)
                    sf_refs.append(ref)
                extras["refs"] = sf_refs
            elif argName == "sourceLine":
                if isinstance(argValue, int):    # must be sortable with int's in logger
                    extras["sourceLine"] = argValue
            elif argName not in ("exc_info", "messageCodes"):
                fmtArgs[argName] = self._loggableValue(argValue) # dereference anything not loggable

        if "refs" not in extras:
            if sourceModelDocument is not None:
                file = sourceModelDocument.displayUri
            else:
                if entryLoadingUrl is not None:
                    file = os.path.basename(entryLoadingUrl)
                else:
                    file = ""
            extras["refs"] = [{"href": file}]
        for pluginXbrlMethod in pluginClassMethods("Logging.Message.Parameters"):
            # plug in can rewrite msg string or return msg if not altering msg
            msg = pluginXbrlMethod(messageCode, msg, modelObjectArgs, fmtArgs) or msg
        return (messageCode,
                (msg, fmtArgs) if fmtArgs else (msg,),
                extras)

    def _loggableValue(self, argValue: Any) -> LoggableValue:  # must be dereferenced and not related to object lifetimes
        if argValue is None:
            return "(none)"
        if isinstance(argValue, bool):
            return str(argValue).lower()  # show lower case true/false xml values
        if isinstance(argValue, int):
            # need locale-dependent formatting
            return format_string(self._modelManager.locale, '%i', argValue)
        if isinstance(argValue, (float, Decimal)):
            # need locale-dependent formatting
            return format_string(self._modelManager.locale, '%f', argValue)
        if isinstance(argValue, tuple):
            return tuple(self._loggableValue(x) for x in argValue)
        if isinstance(argValue, list):
            return [self._loggableValue(x) for x in argValue]
        if isinstance(argValue, set):
            return {self._loggableValue(x) for x in argValue}
        if isinstance(argValue, dict):
            return dict((self._loggableValue(k), self._loggableValue(v)) for k, v in argValue.items())
        return str(argValue)

    def setErrorCaptureLevel(self, errorCaptureLevel: int) -> None:
        self._errorCaptureLevel = errorCaptureLevel



"""
See COPYRIGHT.md for copyright information.
"""
from __future__ import annotations

from dataclasses import dataclass
from typing import Literal


@dataclass(frozen=True)
class ContextRequirement:
    """
    If {elementExists} is set, and {elementDoesNotExist} is not set,
    any context with ID starting with {contextId} must have {element} value
    matching the {elementMatch} DEI value (adjusted by {dayAdjustment} days).
    """
    contextId: str
    element: Literal['endDate', 'instant', 'startDate']
    elementExists: str | None
    elementDoesNotExist: str | None
    elementMatch: str
    dayAdjustment: int = 0  # days to adjust by (e.g. -1 for "day before")


# See section 3-4-2 in the Validation Guidelines.
CONTEXT_REQUIREMENTS = [
    ContextRequirement('CurrentYearInstant', 'instant', None, None, 'CurrentFiscalYearEndDateDEI'),
    ContextRequirement('Prior1YearInstant', 'instant', 'EndDateOfQuarterlyOrSemiAnnualPeriodOfNextFiscalYearDEI', None, 'CurrentFiscalYearEndDateDEI'),
    ContextRequirement('Prior1YearInstant', 'instant', None, 'EndDateOfQuarterlyOrSemiAnnualPeriodOfNextFiscalYearDEI', 'PreviousFiscalYearEndDateDEI'),
    ContextRequirement('Prior2YearInstant', 'instant', 'EndDateOfQuarterlyOrSemiAnnualPeriodOfNextFiscalYearDEI', None, 'PreviousFiscalYearEndDateDEI'),
    ContextRequirement('Prior2YearInstant', 'instant', None, 'EndDateOfQuarterlyOrSemiAnnualPeriodOfNextFiscalYearDEI', 'PreviousFiscalYearStartDateDEI', -1),
    ContextRequirement('CurrentQuarterInstant', 'instant', 'EndDateOfQuarterlyOrSemiAnnualPeriodOfNextFiscalYearDEI', None, 'EndDateOfQuarterlyOrSemiAnnualPeriodOfNextFiscalYearDEI'),
    ContextRequirement('CurrentQuarterInstant', 'instant', None, 'EndDateOfQuarterlyOrSemiAnnualPeriodOfNextFiscalYearDEI', 'CurrentPeriodEndDateDEI'),
    ContextRequirement('Prior1QuarterInstant', 'instant', None, 'EndDateOfQuarterlyOrSemiAnnualPeriodOfNextFiscalYearDEI', 'ComparativePeriodEndDateDEI'),
    ContextRequirement('InterimInstant', 'instant', 'EndDateOfQuarterlyOrSemiAnnualPeriodOfNextFiscalYearDEI', None, 'EndDateOfQuarterlyOrSemiAnnualPeriodOfNextFiscalYearDEI'),
    ContextRequirement('InterimInstant', 'instant', None, 'EndDateOfQuarterlyOrSemiAnnualPeriodOfNextFiscalYearDEI', 'CurrentPeriodEndDateDEI'),
    ContextRequirement('Prior1InterimInstant', 'instant', None, 'EndDateOfQuarterlyOrSemiAnnualPeriodOfNextFiscalYearDEI', 'ComparativePeriodEndDateDEI'),
    ContextRequirement('CurrentYearDuration', 'startDate', None, None, 'CurrentFiscalYearStartDateDEI'),
    ContextRequirement('Prior1YearDuration', 'startDate', 'NextFiscalYearStartDateDEI', None, 'CurrentFiscalYearStartDateDEI'),
    ContextRequirement('Prior1YearDuration', 'startDate', None, 'NextFiscalYearStartDateDEI', 'PreviousFiscalYearStartDateDEI'),
    ContextRequirement('Prior2YearDuration', 'startDate', 'NextFiscalYearStartDateDEI', None, 'PreviousFiscalYearStartDateDEI'),
    ContextRequirement('CurrentYTDDuration', 'startDate', 'NextFiscalYearStartDateDEI', None, 'NextFiscalYearStartDateDEI'),
    ContextRequirement('CurrentYTDDuration', 'startDate', None, 'NextFiscalYearStartDateDEI', 'CurrentFiscalYearStartDateDEI'),
    ContextRequirement('Prior1YTDDuration', 'startDate', None, 'NextFiscalYearStartDateDEI', 'PreviousFiscalYearStartDateDEI'),
    ContextRequirement('InterimDuration', 'startDate', 'NextFiscalYearStartDateDEI', None, 'NextFiscalYearStartDateDEI'),
    ContextRequirement('InterimDuration', 'startDate', None, 'NextFiscalYearStartDateDEI', 'CurrentFiscalYearStartDateDEI'),
    ContextRequirement('Prior1InterimDuration', 'startDate', None, 'NextFiscalYearStartDateDEI', 'PreviousFiscalYearStartDateDEI'),
    ContextRequirement('CurrentYearDuration', 'endDate', None, None, 'CurrentFiscalYearEndDateDEI'),
    ContextRequirement('Prior1YearDuration', 'endDate', 'EndDateOfQuarterlyOrSemiAnnualPeriodOfNextFiscalYearDEI', None, 'CurrentFiscalYearEndDateDEI'),
    ContextRequirement('Prior1YearDuration', 'endDate', None, 'EndDateOfQuarterlyOrSemiAnnualPeriodOfNextFiscalYearDEI', 'PreviousFiscalYearEndDateDEI'),
    ContextRequirement('Prior2YearDuration', 'endDate', 'EndDateOfQuarterlyOrSemiAnnualPeriodOfNextFiscalYearDEI', None, 'PreviousFiscalYearEndDateDEI'),
    ContextRequirement('CurrentYTDDuration', 'endDate', 'EndDateOfQuarterlyOrSemiAnnualPeriodOfNextFiscalYearDEI', None, 'EndDateOfQuarterlyOrSemiAnnualPeriodOfNextFiscalYearDEI'),
    ContextRequirement('CurrentYTDDuration', 'endDate', None, 'EndDateOfQuarterlyOrSemiAnnualPeriodOfNextFiscalYearDEI', 'CurrentPeriodEndDateDEI'),
    ContextRequirement('Prior1YTDDuration', 'endDate', None, 'EndDateOfQuarterlyOrSemiAnnualPeriodOfNextFiscalYearDEI', 'ComparativePeriodEndDateDEI'),
    ContextRequirement('InterimDuration', 'endDate', 'EndDateOfQuarterlyOrSemiAnnualPeriodOfNextFiscalYearDEI', None, 'EndDateOfQuarterlyOrSemiAnnualPeriodOfNextFiscalYearDEI'),
    ContextRequirement('InterimDuration', 'endDate', None, 'EndDateOfQuarterlyOrSemiAnnualPeriodOfNextFiscalYearDEI', 'CurrentPeriodEndDateDEI'),
    ContextRequirement('Prior1InterimDuration', 'endDate', None, 'EndDateOfQuarterlyOrSemiAnnualPeriodOfNextFiscalYearDEI', 'ComparativePeriodEndDateDEI'),
]



from __future__ import annotations

from collections import defaultdict
from typing import (
    TYPE_CHECKING,
    Callable,
    TypeVar,
)

if TYPE_CHECKING:
    from collections.abc import Iterable

T = TypeVar('T')
K = TypeVar('K')

def partitionIntoEquivalenceClasses(items: Iterable[T], key: Callable[[T], K]) -> dict[K, tuple[T, ...]]:
    d = defaultdict(list)
    for item in items:
        d[key(item)].append(item)
    return {k: tuple(v) for k, v in d.items()}



from dataclasses import dataclass

@dataclass
class ConformanceSuiteIdOverride:
    pathSuffix: str
    readMeFirstUris: list[str]
    oldId: str
    newId: str

@dataclass
class ConformanceSuiteIdOverrides:
    pathContainsString: str
    overrides: tuple[ConformanceSuiteIdOverride, ...]

# Some conformance suites reuse variation IDs.
CONFORMANCE_SUITE_ID_OVERRIDES: tuple[ConformanceSuiteIdOverrides, ...] = (
    ConformanceSuiteIdOverrides(
        pathContainsString='conformance-suite-2025-sbr-domein-handelsregister',
        overrides=(
            ConformanceSuiteIdOverride(
                pathSuffix='conformance-suite-2025-sbr-domein-handelsregister/tests/G4-2-2_2/index.xml',
                readMeFirstUris=['TC4_invalid.xbri'],
                oldId='TC3_invalid',
                newId='TC4_invalid',
            ),
            ConformanceSuiteIdOverride(
                pathSuffix='conformance-suite-2025-sbr-domein-handelsregister/tests/G5-1-3_2/index.xml',
                readMeFirstUris=['TC2_valid.xbri'],
                oldId='TC1_valid',
                newId='TC2_valid',
            ),
        ),
    ),
)



from __future__ import annotations

from arelle.ModelDtsObject import ModelConcept
from arelle.ModelXbrl import ModelXbrl
from arelle.utils.validate.Common import isExtensionUri


def getExtensionConcepts(modelXbrl: ModelXbrl, taxonomyUrlPrefixes: frozenset[str]) -> list[ModelConcept]:
    """
    Returns a list of extension concepts in the DTS.
    """
    extensionConcepts = []
    for concepts in modelXbrl.nameConcepts.values():
        for concept in concepts:
            if isExtensionUri(concept.document.uri, modelXbrl, taxonomyUrlPrefixes):
                extensionConcepts.append(concept)
    return extensionConcepts



"""
See COPYRIGHT.md for copyright information.
"""
from __future__ import annotations

import uuid
from collections import defaultdict
from typing import TYPE_CHECKING

from arelle.ModelValue import QName

if TYPE_CHECKING:
    from arelle.ModelInstanceObject import ModelFact


noneUUID = uuid.uuid4()

class FactAspectsCache:
    def __init__(self, maxSize: int) -> None:
        self._maxSize = maxSize if maxSize >= 0 else float("inf")
        self.clear()

    def clear(self) -> None:
        # Dictionaries and sets only undergo resizing upon insertion. Clearing them does not reclaim memory.
        self._size = 0
        self._prioritizedAspects: set[int | QName] = set()
        self._matchingAspects: defaultdict[
            tuple[uuid.UUID, uuid.UUID],
            defaultdict[int | QName, bool | None]
        ] = defaultdict(lambda: defaultdict(lambda: None))

    @property
    def prioritizedAspects(self) -> set[int | QName]:
        return self._prioritizedAspects

    def evaluations(self, fact1: ModelFact, fact2: ModelFact) -> defaultdict[int | QName, bool | None] | None:
        factsCacheKey = self._buildFactKey(fact1, fact2)
        return self._matchingAspects.get(factsCacheKey)

    def cacheMatch(self, fact1: ModelFact, fact2: ModelFact, aspect: int | QName) -> None:
        self._register(fact1, fact2, aspect, True)

    def cacheNotMatch(self, fact1: ModelFact, fact2: ModelFact, aspect: int | QName) -> None:
        self._prioritizedAspects.add(aspect)
        self._register(fact1, fact2, aspect, False)

    def _register(self, fact1: ModelFact, fact2: ModelFact, aspect: int | QName, value: bool) -> None:
        if self._size >= self._maxSize:
            # Stopping additions to the cache entirely is somewhat rudimentary. Alternative caching strategies, such as
            # LRU, were explored, but they demonstrated poorer performance on average across all tested documents. The
            # current strategy involves retaining the cache, but preventing additions once it reaches the maximum size.
            return
        self._size += 1
        factsCacheKey = self._buildFactKey(fact1, fact2)
        self._matchingAspects[factsCacheKey][aspect] = value

    def _buildFactKey(self, fact1: ModelFact, fact2: ModelFact) -> tuple[uuid.UUID, uuid.UUID]:
        fact1Id = fact1.uniqueUUID if fact1 is not None else noneUUID
        fact2Id = fact2.uniqueUUID if fact2 is not None else noneUUID
        return min(fact1Id, fact2Id), max(fact1Id, fact2Id)

    def __repr__(self) -> str:
        return f"FactAspectsCache(size={self._size}, maxSize={self._maxSize}, prioritizedAspects={self._prioritizedAspects}, matchingAspects={self._matchingAspects})"



"""
See COPYRIGHT.md for copyright information.
"""

from typing_extensions import TypeAlias
from arelle.ModelValue import QName

XbrlDtsType: TypeAlias = "XbrlDts"
XbrlLabelType: TypeAlias = "XbrlLabel"
XbrlPropertyType: TypeAlias = "XbrlProperty"
XbrlTaxonomyType: TypeAlias = "XbrlTaxonomy"
class QNameKeyType(QName): # a QName which is also the primary key for parent collection object
    pass
class SQNameKeyType(QName): # an SQName which is also the primary key for parent collection object
    pass
class strKeyType(str): # a str which is also the primary key for parent collection object
    pass
class DefaultTrue: # a bool which if absent defaults to true
    pass
class DefaultFalse: # a bool which if absent defaults to false
    pass
class DefaultZero: # a number which if absent defaults to zero
    pass
